# Copyright 2017 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import uuid

from diskimage_builder.block_device.exception \
    import BlockDeviceSetupException
from diskimage_builder.block_device.plugin import NodeBase
from diskimage_builder.block_device.plugin import PluginBase
from diskimage_builder.block_device.utils import exec_sudo


logger = logging.getLogger(__name__)


# There is the need to check the length of the label of
# the filesystem.  The maximum length depends on the used filesystem.
# This map provides information about the maximum label length.
file_system_max_label_length = {
    "ext2": 16,
    "ext3": 16,
    "ext4": 16,
    "xfs": 12,
    "vfat": 11,
    "swap": 16
}


class FilesystemNode(NodeBase):

    def __init__(self, config, state):
        logger.debug("Create filesystem object; config [%s]", config)
        super(FilesystemNode, self).__init__(config['name'], state)

        # Parameter check (mandatory)
        for pname in ['base', 'type']:
            if pname not in config:
                raise BlockDeviceSetupException(
                    "Mkfs config needs [%s]" % pname)
            setattr(self, pname, config[pname])

        # Parameter check (optional)
        for pname in ['label', 'opts', 'uuid']:
            setattr(self, pname,
                    config[pname] if pname in config else None)

        if self.label is None:
            self.label = self.name

        # for fat/vfat, we use the label as an identifier for the disk
        # so we need that the label is converted to upper case
        if self.type in ('vfat', 'fat'):
            self.label = self.label.upper()

        # ensure we don't already have a fs with this label ... they
        # all must be unique.
        if 'fs_labels' in self.state:
            if self.label in self.state['fs_labels']:
                raise BlockDeviceSetupException(
                    "File system label [%s] used more than once" % self.label)
            self.state['fs_labels'].append(self.label)
        else:
            self.state['fs_labels'] = [self.label]

        if self.type in file_system_max_label_length:
            if file_system_max_label_length[self.type] < len(self.label):
                raise BlockDeviceSetupException(
                    "Label [{label}] too long for filesystem [{type}]: "
                    "{len} > {max_len}".format(**{
                        'label': self.label,
                        'type': self.type,
                        'len': len(self.label),
                        'max_len': file_system_max_label_length[self.type]}))
        else:
            logger.warning("Length of label [%s] cannot be checked for "
                           "filesystem [%s]: unknown max length",
                           self.label, self.type)
            logger.warning("Continue - but this might lead to an error")

        if self.opts is not None:
            self.opts = self.opts.strip().split(' ')

        if self.uuid is None:
            self.uuid = str(uuid.uuid4())

        logger.debug("Filesystem created [%s]", self)

    def get_edges(self):
        edge_from = [self.base]
        edge_to = []
        return (edge_from, edge_to)

    def create(self):
        if self.type in ('swap'):
            cmd = ["mkswap"]
        else:
            cmd = ["mkfs"]
            cmd.extend(['-t', self.type])
            if self.opts:
                cmd.extend(self.opts)

        if self.type in ('vfat', 'fat'):
            cmd.extend(["-n", self.label])
        else:
            cmd.extend(["-L", self.label])

        if self.type in ('ext2', 'ext3', 'ext4', 'swap'):
            cmd.extend(['-U', self.uuid])
        elif self.type == 'xfs':
            cmd.extend(['-m', "uuid=%s" % self.uuid])
        else:
            logger.warning("UUID will not be written for fs type [%s]",
                           self.type)

        if self.type in ('ext2', 'ext3', 'ext4', 'xfs'):
            cmd.append('-q')

        if 'blockdev' not in self.state:
            self.state['blockdev'] = {}
        device = self.state['blockdev'][self.base]['device']
        cmd.append(device)

        logger.debug("Creating fs command [%s]", cmd)
        exec_sudo(cmd)

        if 'filesys' not in self.state:
            self.state['filesys'] = {}
        self.state['filesys'][self.name] \
            = {'uuid': self.uuid, 'label': self.label,
               'fstype': self.type, 'opts': self.opts,
               'device': device}


class Mkfs(PluginBase):
    """Create a file system

    This block device module handles creating different file
    systems.
    """

    def __init__(self, config, defaults, state):
        super(Mkfs, self).__init__()
        self.filesystems = {}
        fs = FilesystemNode(config, state)
        self.filesystems[fs.get_name()] = fs

    def get_nodes(self):
        nodes = []
        for _, fs in self.filesystems.items():
            nodes.append(fs)
        return nodes



# Copyright 2016 Ian Wienand (iwienand@redhat.com)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import os.path
import sys

import diskimage_builder.paths


def main():
    environ = os.environ

    # pre-seed some paths for the shell script
    environ['_LIB'] = diskimage_builder.paths.get_path('lib')

    # export the path to the current python
    if not os.environ.get('DIB_PYTHON_EXEC'):
        os.environ['DIB_PYTHON_EXEC'] = sys.executable

    # we have to handle being called as "disk-image-create" or
    # "ramdisk-image-create".  ramdisk-image-create is just a symlink
    # to disk-image-create
    # XXX: we could simplify things by removing the symlink, and
    # just setting IS_RAMDISK in environ here depending on sys.argv[1]
    script = "%s/%s" % (diskimage_builder.paths.get_path('lib'),
                        os.path.basename(sys.argv[0]))

    os.execve("/bin/bash", ['bash', script] + sys.argv[1:], environ)



# Copyright 2023 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import argparse
import collections
import io
import jsonschema
import os
import os.path
import shlex
import subprocess
import sys
import textwrap
import yaml

import diskimage_builder.paths


class SchemaProperty(object):
    """Base class for a basic schema and a help string"""

    key = None
    description = None
    schema_type = "string"

    def __init__(self, key, description, schema_type=None):
        self.key = key
        self.description = description
        if schema_type:
            self.schema_type = schema_type

    def to_schema(self):
        return {self.key: {"type": self.schema_type}}

    def type_help(self):
        return "Value is a string"

    def to_help(self):
        return "%s\n%s\n%s\n(%s)" % (
            self.key,
            "-" * len(self.key),
            "\n".join(textwrap.wrap(self.description)),
            self.type_help(),
        )


class Env(SchemaProperty):
    """String dict schema for environment variables"""

    def __init__(self, key, description):
        super(Env, self).__init__(key, description, schema_type="object")

    def to_schema(self):
        schema = super(Env, self).to_schema()
        schema[self.key]["additionalProperties"] = {"type": "string"}
        return schema

    def type_help(self):
        return "Value is a map of strings"


class Arg(SchemaProperty):
    """Command argument with associated value"""

    arg = None

    def __init__(self, key, description, schema_type=None, arg=None):
        super(Arg, self).__init__(key, description, schema_type=schema_type)
        self.arg = arg

    def arg_name(self):
        if self.arg is None:
            return "--%s" % self.key
        return self.arg

    def to_argument(self, value=None):
        arg = self.arg_name()
        if value is not None and value != "":
            return [arg, value]
        return []


class Flag(Arg):
    """Boolean value which does not contribute to arguments"""

    def __init__(self, key, description):
        super(Flag, self).__init__(key, description, schema_type="boolean")

    def to_argument(self, value=None):
        return []

    def type_help(self):
        return "Value is a boolean"


class ArgFlag(Arg):
    """Boolean value for a flag argument being set or not"""

    def __init__(self, key, description, arg=None):
        super(ArgFlag, self).__init__(
            key, description, arg=arg, schema_type="boolean"
        )

    def to_argument(self, value=None):
        if value:
            return [self.arg_name()]
        return []

    def type_help(self):
        return "Value is a boolean"


class ArgEnum(Arg):
    """String argument constrained to a list of allowed values"""

    enum = None

    def __init__(
        self, key, description, schema_type="string", arg=None, enum=None
    ):
        super(ArgEnum, self).__init__(
            key, description, schema_type=schema_type, arg=arg
        )
        self.enum = enum and enum or []

    def to_schema(self):
        schema = super(ArgEnum, self).to_schema()
        schema[self.key]["enum"] = self.enum
        return schema

    def type_help(self):
        return "Allowed values: %s" % ", ".join(self.enum)


class ArgFlagRepeating(ArgEnum):
    """Flag argument which repeats the specified number of times"""

    def __init__(self, key, description, arg=None, max_repeat=0):
        enum = list(range(max_repeat + 1))
        super(ArgFlagRepeating, self).__init__(
            key, description, schema_type="integer", arg=arg, enum=enum
        )

    def to_argument(self, value):
        return [self.arg] * value

    def type_help(self):
        return "Allowed values: %s" % ", ".join([str(i) for i in self.enum])


class ArgInt(Arg):
    """Integer argument which a minumum constraint"""

    minimum = 1

    def __init__(self, key, description, arg=None, minimum=1):
        super(ArgInt, self).__init__(
            key, description, arg=arg, schema_type="integer"
        )
        self.minimum = minimum

    def to_schema(self):
        schema = super(ArgInt, self).to_schema()
        schema[self.key]["minimum"] = self.minimum
        return schema

    def to_argument(self, value):
        return super(ArgInt, self).to_argument(str(value))

    def type_help(self):
        return "Value is an integer"


class ArgList(Arg):
    """List of strings converted to comma delimited argument"""

    def __init__(self, key, description, arg=None):
        super(ArgList, self).__init__(
            key, description, arg=arg, schema_type="array"
        )

    def to_schema(self):
        schema = super(ArgList, self).to_schema()
        schema[self.key]["items"] = {"type": "string"}
        return schema

    def to_argument(self, value):
        if not value:
            return []
        return super(ArgList, self).to_argument(",".join(value))

    def type_help(self):
        return "Value is a list of strings"


class ArgListPositional(ArgList):
    """List of strings converted to positional arguments"""

    def __init__(self, key, description):
        super(ArgListPositional, self).__init__(key, description)

    def to_argument(self, value):
        # it is already a list, just return it
        return value

    def type_help(self):
        return "Value is a list of strings"


class ArgEnumList(ArgList):
    """List of strings constrained to a list of allowed values"""

    enum = None

    def __init__(self, key, description, arg=None, enum=None):
        super(ArgEnumList, self).__init__(key, description, arg=arg)
        self.enum = enum and enum or []

    def to_schema(self):
        schema = super(ArgEnumList, self).to_schema()
        schema[self.key]["items"]["enum"] = self.enum
        return schema

    def type_help(self):
        return (
            "Value is a list of strings with allowed values: %s)"
            % ", ".join(self.enum)
        )


class ArgDictToString(Arg):
    """Dict with string values converted to key=value,key2=value2 argument"""

    def __init__(self, key, description, arg=None):
        super(ArgDictToString, self).__init__(
            key, description, arg=arg, schema_type="object"
        )

    def to_schema(self):
        schema = super(ArgDictToString, self).to_schema()
        schema[self.key]["additionalProperties"] = {"type": "string"}
        return schema

    def to_argument(self, value):
        as_list = []
        for k, v in value.items():
            as_list.append("%s=%s" % (k, v))
        return super(ArgDictToString, self).to_argument(",".join(as_list))

    def type_help(self):
        return "Value is a map of strings"


PROPERTIES = [
    Arg("imagename", "Set the imagename of the output image file.", arg="-o"),
    ArgEnum(
        "arch",
        "Set the architecture of the image.",
        arg="-a",
        enum=[
            "aarch64",
            "amd64",
            "arm64",
            "armhf",
            "powerpc",
            "ppc64",
            "ppc64el",
            "ppc64le",
            "s390x",
            "x86_64",
        ],
    ),
    ArgEnumList(
        "types",
        "Set the image types of the output image files.",
        arg="-t",
        enum=[
            "qcow2",
            "tar",
            "tgz",
            "squashfs",
            "vhd",
            "docker",
            "aci",
            "raw",
        ],
    ),
    Env(
        "environment",
        "Environment variables to set during the image build.",
    ),
    Flag(
        "ramdisk",
        "Whether to build a ramdisk image.",
    ),
    ArgFlagRepeating(
        "debug-trace",
        "Tracing level to log, integer 0 is off.",
        arg="-x",
        max_repeat=2,
    ),
    ArgFlag(
        "uncompressed",
        "Do not compress the image - larger but faster.",
        arg="-u",
    ),
    ArgFlag("clear", "Clear environment before starting work.", arg="-c"),
    Arg(
        "logfile",
        "Save run output to given logfile.",
    ),
    ArgFlag(
        "checksum",
        "Generate MD5 and SHA256 checksum files for the created image.",
    ),
    ArgInt(
        "image-size",
        "Image size in GB for the created image.",
    ),
    ArgInt(
        "image-extra-size",
        "Extra image size in GB for the created image.",
    ),
    Arg(
        "image-cache",
        "Location for cached images, defaults to ~/.cache/image-create.",
    ),
    ArgInt(
        "max-online-resize",
        "Max number of filesystem blocks to support when resizing. "
        "Useful if you want a really large root partition when the "
        "image is deployed. Using a very large value may run into a "
        "known bug in resize2fs. Setting the value to 274877906944 "
        "will get you a 1PB root file system. Making this "
        "value unnecessarily large will consume extra disk "
        "space on the root partition with extra file system inodes.",
    ),
    ArgInt(
        "min-tmpfs",
        "Minimum size in GB needed in tmpfs to build the image.",
    ),
    ArgInt(
        "mkfs-journal-size",
        "Filesystem journal size in MB to pass to mkfs.",
    ),
    Arg(
        "mkfs-options",
        "Option flags to be passed directly to mkfs.",
    ),
    ArgFlag("no-tmpfs", "Do not use tmpfs to speed image build."),
    ArgFlag("offline", "Do not update cached resources."),
    ArgDictToString(
        "qemu-img-options",
        "Option flags to be passed directly to qemu-img.",
    ),
    Arg(
        "root-label",
        'Label for the root filesystem, defaults to "cloudimg-rootfs".',
    ),
    Arg(
        "ramdisk-element",
        "Specify the main element to be used for building ramdisks. "
        'Defaults to "ramdisk".  Should be set to "dracut-ramdisk" '
        "for platforms such as RHEL and CentOS that do not package busybox.",
    ),
    ArgEnum(
        "install-type",
        "Specify the default installation type.",
        enum=["source", "package"],
    ),
    Arg(
        "docker-target",
        "Specify the repo and tag to use if the output type is docker, "
        "defaults to the value of output imagename.",
    ),
    ArgList(
        "packages",
        "Extra packages to install in the image.  Runs once, after "
        '"install.d" phase. Does not apply when ramdisk is true.',
        arg="-p",
    ),
    ArgFlag(
        "skip-base",
        'Skip the default inclusion of the "base" element. '
        "Does not apply when ramdisk is true.",
        arg="-n",
    ),
    ArgListPositional(
        "elements",
        "list of elements to build the image with",
    ),
]


SCHEMA_PROPERTIES = {}
for arg in PROPERTIES:
    SCHEMA_PROPERTIES.update(arg.to_schema())

DIB_SCHEMA = {
    "type": "array",
    "items": {
        "type": "object",
        "properties": SCHEMA_PROPERTIES,
        "additionalProperties": False,
    },
    "additionalProperties": False,
}


class Command(object):
    script = None
    args = None
    environ = None

    def __init__(self, script, properties, entry):
        self.script = script
        self.args = []
        self.environ = {}
        for prop in properties:
            if prop.key in entry:
                value = entry[prop.key]
                if isinstance(prop, Env):
                    self.environ.update(value)
                elif isinstance(prop, Arg):
                    self.args.extend(prop.to_argument(value))

    def merged_env(self):
        environ = os.environ.copy()
        # pre-seed some paths for the shell script
        environ["_LIB"] = diskimage_builder.paths.get_path("lib")
        environ.update(self.environ)
        return environ

    def command(self):
        return ["bash", self.script] + self.args

    def __repr__(self):
        elements = []
        for k, v in self.environ.items():
            elements.append("%s=%s" % (k, shlex.quote(v)))
        elements.extend([shlex.quote(a) for a in self.command()])
        return " ".join(elements) + "\n"


def help_properties():
    str = io.StringIO()
    for prop in PROPERTIES:
        str.write(prop.to_help())
        str.write("\n\n")
    return str.getvalue()


def get_args():
    description = (
        """\
The file format is YAML which expects a list of image definition maps.

Supported entries for an image definition are:

%s
"""
        % help_properties()
    )
    parser = argparse.ArgumentParser(
        description=description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "files",
        metavar="<filename>",
        nargs="+",
        help="Paths to image build definition YAML files",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        help="Show the disk-image-create, ramdisk-image-create commands and "
        "exit",
    )
    parser.add_argument(
        "--stop-on-failure",
        action="store_true",
        help="Stop building images when an image build fails",
    )
    args = parser.parse_args(sys.argv[1:])
    return args


def merge_entry(merged_entry, entry):
    for k, v in entry.items():
        if isinstance(v, list):
            # append to existing list
            list_value = merged_entry.setdefault(k, [])
            list_value.extend(v)
        elif isinstance(v, dict):
            # update environment dict
            dict_value = merged_entry.setdefault(k, {})
            dict_value.update(v)
        else:
            # update value
            merged_entry[k] = v


def build_commands(definition):
    jsonschema.validate(definition, schema=DIB_SCHEMA)
    dib_script = "%s/disk-image-create" % diskimage_builder.paths.get_path(
        "lib"
    )
    rib_script = "%s/ramdisk-image-create" % diskimage_builder.paths.get_path(
        "lib"
    )

    # Start with the default image name, 'image'
    previous_imagename = "image"
    merged_entries = collections.OrderedDict()
    for entry in definition:
        imagename = entry.get("imagename", previous_imagename)
        previous_imagename = imagename
        if imagename not in merged_entries:
            merged_entries[imagename] = entry
        else:
            merge_entry(merged_entries[imagename], entry)

    commands = []
    for entry in merged_entries.values():
        if entry.get("ramdisk", False):
            commands.append(Command(rib_script, PROPERTIES, entry))
        else:
            commands.append(Command(dib_script, PROPERTIES, entry))
    return commands


def main():
    args = get_args()

    # export the path to the current python
    if not os.environ.get("DIB_PYTHON_EXEC"):
        os.environ["DIB_PYTHON_EXEC"] = sys.executable

    definitions = []
    for file in args.files:
        with open(file) as f:
            definitions.extend(yaml.safe_load(f))
    commands = build_commands(definitions)
    final_returncode = 0
    failed_command = None
    for command in commands:
        sys.stderr.write(str(command))
        sys.stderr.write("\n")
        sys.stderr.flush()
        if not args.dry_run:
            p = subprocess.Popen(command.command(), env=command.merged_env())
            p.communicate()
            if p.returncode != 0:
                final_returncode = p.returncode
                failed_command = command
                if args.stop_on_failure:
                    break

    if final_returncode != 0:
        raise subprocess.CalledProcessError(
            final_returncode, failed_command.command()
        )



# Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import glob
import os


import testtools


class TestNoDupFilenames(testtools.TestCase):

    def test_no_dup_filenames(self):
        topdir = os.path.normpath(os.path.dirname(__file__) + '/../../')
        elements_glob = os.path.join(topdir, "elements", "*")

        filenames = []
        dirs_to_check = ['block-device.d', 'cleanup.d', 'extra-data.d',
                         'finalise.d', 'install.d', 'post-install.d',
                         'pre-install.d', 'root.d']

        for element_dir in glob.iglob(elements_glob):
            for dir_to_check in dirs_to_check:
                target_dir = os.path.join(element_dir, dir_to_check, "*")
                for target in glob.iglob(target_dir):
                    short_path = target[len(element_dir) + 1:]
                    if not os.path.isdir(target):
                        err_msg = 'Duplicate file name found %s' % short_path
                        self.assertFalse(short_path in filenames, err_msg)
                        filenames.append(short_path)



# Copyright 2017 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from diskimage_builder.block_device.plugin import NodeBase
from diskimage_builder.block_device.plugin import PluginBase


logger = logging.getLogger(__name__)


class FstabNode(NodeBase):
    def __init__(self, config, state):
        super(FstabNode, self).__init__(config['name'], state)
        self.base = config['base']
        self.options = config.get('options', 'defaults')
        self.dump_freq = config.get('dump-freq', 0)
        self.fsck_passno = config.get('fsck-passno', 2)

    def get_edges(self):
        edge_from = [self.base]
        edge_to = []
        return (edge_from, edge_to)

    def create(self):
        logger.debug("fstab create called [%s]", self.name)

        if 'fstab' not in self.state:
            self.state['fstab'] = {}

        swap = 'none' in self.state['mount'] and \
            self.state['mount']['none']['name'] == self.base

        self.state['fstab'][self.base] = {
            'name': self.name,
            'base': self.base,
            'options': 'sw' if swap else self.options,
            'dump-freq': 0 if swap else self.dump_freq,
            'fsck-passno': 0 if swap else self.fsck_passno
        }


class Fstab(PluginBase):
    def __init__(self, config, defaults, state):
        super(Fstab, self).__init__()

        self.node = FstabNode(config, state)

    def get_nodes(self):
        return [self.node]



# Copyright 2023 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import jsonschema
import mock
import subprocess
import tempfile
import testtools
import os
import yaml

from diskimage_builder import diskimage_builder as dib


class TestDib(testtools.TestCase):
    def assert_to_argument(self, expected, prop, value):
        self.assertEqual(expected, prop.to_argument(value))
        self.assert_schema_validate(prop, value)

    def assert_schema_validate(self, prop, value, assert_failure=False):
        # create a fake dict value and validate the schema against it
        value_dict = {prop.key: value}
        schema = {"type": "object", "properties": {}}
        schema["properties"].update(prop.to_schema())
        if assert_failure:
            self.assertRaises(
                jsonschema.exceptions.ValidationError,
                jsonschema.validate,
                value_dict,
                schema=schema,
            )
        else:
            jsonschema.validate(value_dict, schema=schema)

    def test_schema_property(self):
        x = dib.SchemaProperty(
            "the_key", "the description", schema_type="integer"
        )
        self.assertEqual(
            """\
the_key
-------
the description
(Value is a string)""",
            x.to_help(),
        )
        self.assertEqual({"the_key": {"type": "integer"}}, x.to_schema())

    def test_env(self):
        x = dib.Env(
            "environment",
            "environment variables to set during the image build",
        )
        self.assertEqual(
            {
                "environment": {
                    "additionalProperties": {"type": "string"},
                    "type": "object",
                }
            },
            x.to_schema(),
        )

    def test_arg(self):
        # no arg
        x = dib.Arg("the-key", "")
        self.assert_to_argument(["--the-key", "the value"], x, "the value")

        # with arg
        x = dib.Arg("the-key", "", arg="--key")
        self.assert_to_argument(["--key", "the value"], x, "the value")

        # with empty string value
        self.assert_to_argument([], x, "")

    def test_arg_flag(self):
        x = dib.ArgFlag("doit", "do it", arg="-d")
        self.assertEqual({"doit": {"type": "boolean"}}, x.to_schema())

        # false value
        self.assert_to_argument([], x, False)

        # true value
        self.assert_to_argument(["-d"], x, True)

    def test_arg_enum(self):
        x = dib.ArgEnum("choice", "", enum=["one", "two", "three"])
        self.assertEqual(
            {"choice": {"type": "string", "enum": ["one", "two", "three"]}},
            x.to_schema(),
        )
        self.assert_to_argument(["--choice", "one"], x, "one")
        self.assert_schema_validate(x, "two")
        self.assert_schema_validate(x, "four", assert_failure=True)

    def test_arg_flag_repeating(self):
        x = dib.ArgFlagRepeating("log_level", "", arg="-v", max_repeat=3)
        self.assertEqual(
            {"log_level": {"type": "integer", "enum": [0, 1, 2, 3]}},
            x.to_schema(),
        )
        self.assert_to_argument([], x, 0)
        self.assert_to_argument(["-v"], x, 1)
        self.assert_to_argument(["-v", "-v"], x, 2)
        self.assert_to_argument(["-v", "-v", "-v"], x, 3)

    def test_arg_int(self):
        x = dib.ArgInt("size", "")
        self.assertEqual(
            {"size": {"type": "integer", "minimum": 1}}, x.to_schema()
        )
        self.assert_to_argument(["--size", "10"], x, 10)
        self.assert_schema_validate(x, -1, assert_failure=True)

    def test_arg_list(self):
        x = dib.ArgList("packages", "", arg="-p")
        self.assertEqual(
            {"packages": {"type": "array", "items": {"type": "string"}}},
            x.to_schema(),
        )
        self.assert_to_argument(
            ["-p", "wget,vim-enhanced"], x, ["wget", "vim-enhanced"]
        )
        self.assert_to_argument([], x, [])

    def test_arg_list_position(self):
        x = dib.ArgListPositional("elements", "")
        self.assert_to_argument(
            ["centos", "vm", "bootloader"], x, ["centos", "vm", "bootloader"]
        )

    def test_arg_enum_list(self):
        x = dib.ArgEnumList(
            "types", "", arg="-t", enum=["qcow2", "tar", "raw"]
        )
        self.assertEqual(
            {
                "types": {
                    "items": {
                        "enum": ["qcow2", "tar", "raw"],
                        "type": "string",
                    },
                    "type": "array",
                }
            },
            x.to_schema(),
        )
        self.assert_to_argument(["-t", "qcow2,raw"], x, ["qcow2", "raw"])

    def test_arg_dict_to_string(self):
        x = dib.ArgDictToString("options", "")
        self.assertEqual(
            {
                "options": {
                    "additionalProperties": {"type": "string"},
                    "type": "object",
                }
            },
            x.to_schema(),
        )
        self.assert_to_argument(
            ["--options", "foo=bar,baz=ingo"],
            x,
            {"foo": "bar", "baz": "ingo"},
        )

    def test_command_merged_env(self):
        c = dib.Command(
            "echo",
            properties=[dib.Env("environment", "")],
            entry={
                "environment": {
                    "ELEMENTS_PATH": "/path/to/elements",
                    "DIB_CLOUD_IMAGES": "/path/to/image.qcow2",
                }
            },
        )
        env = c.merged_env()
        self.assertEqual("/path/to/elements", env["ELEMENTS_PATH"])
        self.assertEqual("/path/to/image.qcow2", env["DIB_CLOUD_IMAGES"])
        self.assertIn(needle="_LIB", haystack=env)
        # this will be merged with the whole environment, so there will be
        # more than 3 values
        self.assertGreater(len(env), 3)

    def test_command(self):
        c = dib.Command(
            "echo",
            properties=[
                dib.Env("environment", ""),
                dib.Arg("the-key", "", arg="--key"),
                dib.ArgFlag("doit", "do it", arg="-d"),
                dib.ArgFlagRepeating("verbose", "", arg="-v", max_repeat=3),
                dib.ArgDictToString("options", ""),
                dib.ArgListPositional("elements", ""),
            ],
            entry={
                "environment": {
                    "ELEMENTS_PATH": "/path/to/elements",
                    "DIB_CLOUD_IMAGES": "~/image.qcow2",
                },
                "the-key": "the-value",
                "doit": True,
                "verbose": 3,
                "options": {"foo": "bar"},
                "elements": ["centos", "vm"],
            },
        )
        self.assertEqual(
            [
                "bash",
                "echo",
                "--key",
                "the-value",
                "-d",
                "-v",
                "-v",
                "-v",
                "--options",
                "foo=bar",
                "centos",
                "vm",
            ],
            c.command(),
        )
        self.assertEqual(
            "ELEMENTS_PATH=/path/to/elements "
            "DIB_CLOUD_IMAGES='~/image.qcow2' "
            "bash echo --key the-value -d -v -v -v --options foo=bar "
            "centos vm\n",
            str(c),
        )

    def test_merge_entry(self):
        # override normal attributes
        merged_entry = {
            "imagename": "image1",
            "elements": ["one", "two", "three"],
            "debug-trace": 1,
        }
        dib.merge_entry(
            merged_entry,
            {
                "imagename": "image1",
                "debug-trace": 2,
                "logfile": "image1.log",
            },
        )
        self.assertEqual(
            {
                "imagename": "image1",
                "elements": ["one", "two", "three"],
                "debug-trace": 2,
                "logfile": "image1.log",
            },
            merged_entry,
        )

        # append list attributes, update dict attributes
        merged_entry = {
            "imagename": "image1",
            "elements": ["one", "two", "three"],
            "environment": {
                "DIB_ONE": "1",
                "DIB_TWO": "2",
            },
        }
        dib.merge_entry(
            merged_entry,
            {
                "imagename": "image1",
                "elements": ["four", "five"],
                "environment": {
                    "DIB_TWO": "two",
                    "DIB_THREE": "three",
                },
            },
        )
        self.assertEqual(
            {
                "imagename": "image1",
                "elements": ["one", "two", "three", "four", "five"],
                "environment": {
                    "DIB_ONE": "1",
                    "DIB_TWO": "two",
                    "DIB_THREE": "three",
                },
            },
            merged_entry,
        )

    @mock.patch("diskimage_builder.paths.get_path")
    def test_build_commands_simple(self, mock_get_path):
        mock_get_path.return_value = "/lib"
        commands = dib.build_commands(
            [
                {
                    "imagename": "centos-minimal",
                    "elements": ["centos", "vm"],
                },
                {
                    "imagename": "ironic-python-agent",
                    "ramdisk": True,
                    "elements": [
                        "ironic-python-agent-ramdisk",
                        "extra-hardware",
                    ],
                },
            ]
        )

        self.assertEqual(
            [
                "bash",
                "/lib/disk-image-create",
                "-o",
                "centos-minimal",
                "centos",
                "vm",
            ],
            commands[0].command(),
        )
        self.assertEqual(
            [
                "bash",
                "/lib/ramdisk-image-create",
                "-o",
                "ironic-python-agent",
                "ironic-python-agent-ramdisk",
                "extra-hardware",
            ],
            commands[1].command(),
        )

    @mock.patch("diskimage_builder.paths.get_path")
    def test_build_commands_merged(self, mock_get_path):
        mock_get_path.return_value = "/lib"
        commands = dib.build_commands(
            [
                {  # base definition
                    "imagename": "centos-minimal",
                    "elements": ["centos", "vm"],
                    "environment": {"foo": "bar", "zip": "zap"},
                },
                {  # merge with previous when no imagename
                    "elements": ["devuser"],
                    "logfile": "centos-minimal.log",
                },
                {  # merge when same imagename
                    "imagename": "centos-minimal",
                    "environment": {"foo": "baz", "one": "two"},
                },
            ]
        )

        self.assertEqual(1, len(commands))
        self.assertEqual(
            [
                "bash",
                "/lib/disk-image-create",
                "-o",
                "centos-minimal",
                "--logfile",
                "centos-minimal.log",
                "centos",
                "vm",
                "devuser",
            ],
            commands[0].command(),
        )

    @mock.patch("diskimage_builder.paths.get_path")
    def test_build_commands_all_arguments(self, mock_get_path):
        mock_get_path.return_value = "/lib"
        commands = dib.build_commands(
            [
                {
                    "arch": "amd64",
                    "imagename": "everyoption",
                    "types": ["qcow2"],
                    "debug-trace": 2,
                    "uncompressed": True,
                    "clear": True,
                    "logfile": "./logfile.log",
                    "checksum": True,
                    "image-size": 40,
                    "image-extra-size": 1,
                    "image-cache": "~/.cache/dib",
                    "max-online-resize": 1000,
                    "min-tmpfs": 7,
                    "mkfs-journal-size": 1,
                    "mkfs-options": "-D",
                    "no-tmpfs": True,
                    "offline": True,
                    "qemu-img-options": {"size": "10"},
                    "ramdisk-element": "dracut-ramdisk",
                    "install-type": "package",
                    "root-label": "root",
                    "docker-target": "everyoption:latest",
                    "packages": ["wget", "tmux"],
                    "skip-base": True,
                    "elements": ["centos", "vm"],
                }
            ]
        )

        self.assertEqual(
            [
                "bash",
                "/lib/disk-image-create",
                "-o",
                "everyoption",
                "-a",
                "amd64",
                "-t",
                "qcow2",
                "-x",
                "-x",
                "-u",
                "-c",
                "--logfile",
                "./logfile.log",
                "--checksum",
                "--image-size",
                "40",
                "--image-extra-size",
                "1",
                "--image-cache",
                "~/.cache/dib",
                "--max-online-resize",
                "1000",
                "--min-tmpfs",
                "7",
                "--mkfs-journal-size",
                "1",
                "--mkfs-options",
                "-D",
                "--no-tmpfs",
                "--offline",
                "--qemu-img-options",
                "size=10",
                "--root-label",
                "root",
                "--ramdisk-element",
                "dracut-ramdisk",
                "--install-type",
                "package",
                "--docker-target",
                "everyoption:latest",
                "-p",
                "wget,tmux",
                "-n",
                "centos",
                "vm",
            ],
            commands[0].command(),
        )

    def write_image_definition(self):
        image_def = [
            {
                "imagename": "centos-minimal",
                "elements": ["centos", "vm"],
            },
            {
                "imagename": "ironic-python-agent",
                "ramdisk": True,
                "elements": [
                    "ironic-python-agent-ramdisk",
                    "extra-hardware",
                ],
            },
        ]
        with tempfile.NamedTemporaryFile(delete=False) as deffile:
            self.addCleanup(os.remove, deffile.name)
            self.filelist = [deffile.name]
            with open(deffile.name, "w") as f:
                f.write(yaml.dump(image_def))
        return deffile.name

    @mock.patch("diskimage_builder.paths.get_path")
    @mock.patch("diskimage_builder.diskimage_builder.get_args")
    @mock.patch("subprocess.Popen")
    def test_main_dry_run(self, mock_popen, mock_get_args, mock_get_path):
        mock_get_path.return_value = "/lib"
        mock_get_args.return_value = mock.Mock(
            dry_run=True, files=[self.write_image_definition()]
        )
        dib.main()
        mock_popen.assert_not_called()

    @mock.patch("diskimage_builder.paths.get_path")
    @mock.patch("diskimage_builder.diskimage_builder.get_args")
    @mock.patch("subprocess.Popen")
    def test_main(self, mock_popen, mock_get_args, mock_get_path):
        mock_get_path.return_value = "/lib"
        mock_get_args.return_value = mock.Mock(
            dry_run=False,
            files=[self.write_image_definition()],
            stop_on_failure=False,
        )

        process = mock.Mock()
        process.returncode = 0
        mock_popen.return_value = process

        dib.main()
        self.assertEqual(2, mock_popen.call_count)
        mock_popen.assert_has_calls(
            [
                mock.call(
                    [
                        "bash",
                        "/lib/disk-image-create",
                        "-o",
                        "centos-minimal",
                        "centos",
                        "vm",
                    ],
                    env=mock.ANY,
                ),
                mock.call(
                    [
                        "bash",
                        "/lib/ramdisk-image-create",
                        "-o",
                        "ironic-python-agent",
                        "ironic-python-agent-ramdisk",
                        "extra-hardware",
                    ],
                    env=mock.ANY,
                ),
            ],
            any_order=True,
        )

    @mock.patch("diskimage_builder.paths.get_path")
    @mock.patch("diskimage_builder.diskimage_builder.get_args")
    @mock.patch("subprocess.Popen")
    def test_main_stop_on_failure(
        self, mock_popen, mock_get_args, mock_get_path
    ):
        mock_get_path.return_value = "/lib"
        mock_get_args.return_value = mock.Mock(
            dry_run=False,
            files=[self.write_image_definition()],
            stop_on_failure=True,
        )

        process = mock.Mock()
        process.returncode = -1
        mock_popen.return_value = process

        e = self.assertRaises(subprocess.CalledProcessError, dib.main)
        self.assertEqual(1, mock_popen.call_count)
        self.assertEqual(-1, e.returncode)

    @mock.patch("diskimage_builder.paths.get_path")
    @mock.patch("diskimage_builder.diskimage_builder.get_args")
    @mock.patch("subprocess.Popen")
    def test_main_continue_on_failure(
        self, mock_popen, mock_get_args, mock_get_path
    ):
        mock_get_path.return_value = "/lib"
        mock_get_args.return_value = mock.Mock(
            dry_run=False,
            files=[self.write_image_definition()],
            stop_on_failure=False,
        )

        process = mock.Mock()
        process.returncode.side_effect = -1
        mock_popen.return_value = process

        self.assertRaises(subprocess.CalledProcessError, dib.main)
        self.assertEqual(2, mock_popen.call_count)



# Copyright 2017 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging

from diskimage_builder.block_device.exception \
    import BlockDeviceSetupException
from diskimage_builder.block_device.plugin import NodeBase
from diskimage_builder.block_device.plugin import PluginBase
from diskimage_builder.block_device.utils import exec_sudo
from diskimage_builder.block_device.utils import parse_abs_size_spec
from diskimage_builder.block_device.utils import remove_device

PHYSICAL_EXTENT_BYTES = parse_abs_size_spec('4MiB')
LVS_TYPES = ['thin', 'thin-pool']

logger = logging.getLogger(__name__)

#
# LVM
# ---
#
# The LVM config has three required keys; pvs, vgs and lvs
#
# lvm:     ->  LVSNode
#   pvs:   ->  PvsNode
#   lvs:   ->  LvsNode
#   vgs:   ->  VgsNode
#
# The LVMPlugin will verify this and build nodes into the
# configuration graph.
#
# As described below, a LVSNode is created for synchronisation
# purposes.  Thus if you had something like two partitions that became
# two physical-volumes (pv1 & pv2), that you then combine into a
# single volume group (vg) and then create several logical volumes
# (lv1, lv2, lv3) your graph would end up looking like:
#
#  partition1     partition2
#      |               |
#      ---> LVSNode <--+
#              |
#       +------+------+
#       v             v
#      pv1           pv2
#       |             |
#       +-->  vg  <---+
#             |
#       +-----+-----+
#       v     v     v
#      lv1   lv2   lv3
#
# After the create() call on the LVSNode object, the entire LVM setup
# would actually be complete.  The other nodes are all just
# place-holders, and are used for further ordering (for example, the
# fs creation & mounting should depend on the logical volume nodes).
# For this reason, their create() calls are blank.  However, for code
# organisational purposes they have a private _create() and _cleanup()
# call that is driven by the LVSNode object.


class PvsNode(NodeBase):
    def __init__(self, name, state, base, options):
        """Physical volume

        This is a placeholder node for the LVM physical volumes.

        Arguments:
        :param name: Name of this node
        :param state: global state pointer
        :param base: Parent partition
        :param options: config options
        """
        super(PvsNode, self).__init__(name, state)
        self.base = base
        self.options = options

    def _create(self):
        # the underlying device path of our parent was previously
        # recorded into the state during blockdev creation; look it
        # up.
        phys_dev = self.state['blockdev'][self.base]['device']

        cmd = ["pvcreate"]
        cmd.append(phys_dev)
        if self.options:
            cmd.extend(self.options)
        logger.debug("Creating pv command [%s]", cmd)
        exec_sudo(cmd)

        # save state
        if 'pvs' not in self.state:
            self.state['pvs'] = {}
        self.state['pvs'][self.name] = {
            'opts': self.options,
            'device': phys_dev
        }

    def get_edges(self):
        # See LVMNode.get_edges() for how this gets connected
        return ([], [])

    def create(self):
        # see notes in LVMNode object
        pass


class VgsNode(NodeBase):
    def __init__(self, name, state, base, options):
        """Volume Group

        This is a placeholder node for a volume group

        Arguments:
        :param name: Name of this node
        :param state: global state pointer
        :param base: Parent :class:`PvsNodes` this volume group exists on
        :param options: extra options passed to the `vgcreate` command
        """
        super(VgsNode, self).__init__(name, state)
        self.base = base
        self.options = options

    def _create(self):
        # The PV's have saved their actual device name into the state
        # during their _create().  Look at our base elements and thus
        # find the underlying device paths in the state.
        pvs_devs = []
        for pv in self.base:
            pvs_dev = self.state['pvs'][pv]['device']
            pvs_devs.append(pvs_dev)

        cmd = ["vgcreate", ]
        cmd.append(self.name)
        cmd.extend(pvs_devs)
        if self.options:
            cmd.extend(self.options)

        logger.debug("Creating vg command [%s]", cmd)
        exec_sudo(cmd)

        # save state
        if 'vgs' not in self.state:
            self.state['vgs'] = {}
        self.state['vgs'][self.name] = {
            'opts': self.options,
            'devices': self.base,
        }

    def _umount(self):
        exec_sudo(['vgchange', '-an', self.name])

    def get_edges(self):
        # self.base is already a list, per the config.  There might be
        # multiple pv parents here.
        edge_from = self.base
        edge_to = []
        return (edge_from, edge_to)

    def create(self):
        # see notes in LVMNode object
        pass


class LvsNode(NodeBase):
    def __init__(self, name, state, base, options, size, extents, segtype,
                 thin_pool):
        """Logical Volume

        This is a placeholder node for a logical volume

        Arguments:
        :param name: Name of this node
        :param state: global state pointer
        :param base: the parent volume group
        :param options: options passed to lvcreate
        :param size: size of the LV, using the supported unit types
                     (MB, MiB, etc)
        :param extents: size of the LV in extents
        :param segtype: value passed to segment type, supports thin and
                        thin-pool
        :param thin_pool: name of the thin pool to create this volume from
        """
        super(LvsNode, self).__init__(name, state)
        self.base = base
        self.options = options
        self.size = size
        self.extents = extents
        self.type = segtype
        self.thin_pool = thin_pool

    def _create(self):
        cmd = ["lvcreate", ]
        cmd.extend(['--name', self.name])
        if self.type:
            cmd.extend(['--type', self.type])
        if self.thin_pool:
            cmd.extend(['--thin-pool', self.thin_pool])
        if self.size:
            size = parse_abs_size_spec(self.size)
            # ensuire size aligns with physical extents
            size = size - size % PHYSICAL_EXTENT_BYTES
            size_arg = '-V' if self.type == 'thin' else '-L'
            cmd.extend([size_arg, '%dB' % size])
        elif self.extents:
            cmd.extend(['-l', self.extents])
        if self.options:
            cmd.extend(self.options)

        cmd.append(self.base)

        logger.debug("Creating lv command [%s]", cmd)
        exec_sudo(cmd)

        # save state
        device_name = '%s-%s' % (self.base, self.name)
        self.state['blockdev'][self.name] = {
            'vgs': self.base,
            'size': self.size,
            'extents': self.extents,
            'opts': self.options,
            'device': '/dev/mapper/%s' % device_name
        }
        self.add_rollback(remove_device, device_name)

    def _umount(self):
        exec_sudo(['lvchange', '-an',
                   '/dev/%s/%s' % (self.base, self.name)])

    def get_edges(self):
        edge_from = [self.base]
        edge_to = []
        return (edge_from, edge_to)

    def create(self):
        # see notes in LVMNode object
        pass


class LVMNode(NodeBase):
    def __init__(self, name, state, pvs, lvs, vgs):
        """LVM Driver Node

        This is the "global" node where all LVM operations are driven
        from.  In the node graph, the LVM physical-volumes depend on
        this node.  This node then depends on the devices that the
        PV's require.  This node incorporates *all* LVM setup;
        i.e. after the create() call here we have created all pv's,
        lv's and vg.  The <Pvs|Lvs|Vgs>Node objects in the graph are
        therefore just dependency place holders whose create() call
        does nothing.

        Arguments:
        :param name: name of this node
        :param state: global state pointer
        :param pvs: A list of :class:`PvsNode` objects
        :param lvs: A list of :class:`LvsNode` objects
        :param vgs: A list of :class:`VgsNode` objects

        """
        super(LVMNode, self).__init__(name, state)
        self.pvs = pvs
        self.lvs = lvs
        self.vgs = vgs

    def get_edges(self):
        # This node requires the physical device(s), which is
        # recorded in the "base" argument of the PV nodes.
        pvs = []
        for pv in self.pvs:
            pvs.append(pv.base)
        edge_from = set(pvs)

        # The PV nodes should then depend on us.  i.e., we just made
        # this node a synchronisation point
        edge_to = [pv.name for pv in self.pvs]

        return (edge_from, edge_to)

    def create(self):
        # Run through pvs->vgs->lvs and create them
        # XXX: we could theoretically get this same info from walking
        # the graph of our children nodes?  Would that be helpful in
        # any way?
        for pvs in self.pvs:
            pvs._create()
        for vgs in self.vgs:
            vgs._create()
        for lvs in self.lvs:
            lvs._create()

    def umount(self):
        for lvs in self.lvs:
            lvs._umount()

        for vgs in self.vgs:
            vgs._umount()

        exec_sudo(['udevadm', 'settle'])

    def cleanup(self):
        # Information about the PV, VG and LV is typically
        # cached in lvmetad. Even after removing PV device and
        # partitions this data is not automatically updated
        # which leads to a couple of problems.
        # the 'pvscan --cache' scans the available disks
        # and updates the cache.
        # This is in cleanup because it must be called after the
        # umount of the containing block device is done, (which should
        # all be done in umount phase).
        try:
            exec_sudo(['pvscan', '--cache'])
        except BlockDeviceSetupException as e:
            logger.info("pvscan call failed [%s]", e.returncode)


class LVMPlugin(PluginBase):

    def _config_error(self, msg):
        raise BlockDeviceSetupException(msg)

    def __init__(self, config, defaults, state):
        """Build LVM nodes

        This reads the "lvm:" config stanza, validates it and produces
        the PV, VG and LV nodes.  These are all synchronised via a
        LVMNode as described above.

        Arguments:
        :param config: "lvm" configuration dictionary
        :param defaults: global defaults dictionary
        :param state: global state reference
        """

        super(LVMPlugin, self).__init__()

        # note lvm: doesn't require a base ... the base is the
        # physical devices the "pvs" nodes are made on.
        if 'name' not in config:
            self._config_error("Lvm config requires 'name'")
        if 'pvs' not in config:
            self._config_error("Lvm config requires a 'pvs'")
        if 'vgs' not in config:
            self._config_error("Lvm config needs 'vgs'")
        if 'lvs' not in config:
            self._config_error("Lvm config needs 'lvs'")

        # create physical volume nodes
        self.pvs = []
        self.pvs_keys = []
        for pvs_cfg in config['pvs']:
            if 'name' not in pvs_cfg:
                self._config_error("Missing 'name' in pvs config")
            if 'base' not in pvs_cfg:
                self._config_error("Missing 'base' in pvs_config")

            pvs_item = PvsNode(pvs_cfg['name'], state,
                               pvs_cfg['base'],
                               pvs_cfg.get('options'))
            self.pvs.append(pvs_item)

        # create volume group nodes
        self.vgs = []
        self.vgs_keys = []
        for vgs_cfg in config['vgs']:
            if 'name' not in vgs_cfg:
                self._config_error("Missing 'name' in vgs config")
            if 'base' not in vgs_cfg:
                self._config_error("Missing 'base' in vgs config")

            # Ensure we have a valid PVs backing this VG
            for pvs in vgs_cfg['base']:
                if not any(pv.name == pvs for pv in self.pvs):
                    self._config_error("base:%s in vgs does not "
                                       "match a valid pvs" % pvs)
            vgs_item = VgsNode(vgs_cfg['name'], state, vgs_cfg['base'],
                               vgs_cfg.get('options', None))
            self.vgs.append(vgs_item)

        # create logical volume nodes
        self.lvs = []
        for lvs_cfg in config['lvs']:
            if 'name' not in lvs_cfg:
                self._config_error("Missing 'name' in lvs config")
            if 'base' not in lvs_cfg:
                self._config_error("Missing 'base' in lvs config")
            if 'size' not in lvs_cfg and 'extents' not in lvs_cfg:
                self._config_error("Missing 'size' or 'extents' in lvs config")

            # ensure this logical volume has a valid volume group base
            if not any(vg.name == lvs_cfg['base'] for vg in self.vgs):
                self._config_error("base:%s in lvs does not match a valid vg" %
                                   lvs_cfg['base'])

            if 'type' in lvs_cfg:
                if lvs_cfg['type'] not in LVS_TYPES:
                    self._config_error(
                        "Unsupported type:%s, supported types: %s" %
                        (lvs_cfg['type'], ', '.join(LVS_TYPES)))

            lvs_item = LvsNode(lvs_cfg['name'], state, lvs_cfg['base'],
                               lvs_cfg.get('options', None),
                               lvs_cfg.get('size', None),
                               lvs_cfg.get('extents', None),
                               lvs_cfg.get('type', None),
                               lvs_cfg.get('thin-pool', None))
            self.lvs.append(lvs_item)

        # create the "driver" node
        self.lvm_node = LVMNode(config['name'], state,
                                self.pvs, self.lvs, self.vgs)

    def get_nodes(self):
        # the nodes for insertion into the graph are all of the pvs,
        # vgs and lvs nodes we have created above and the root node and
        # the cleanup node.
        return self.pvs + self.vgs + self.lvs \
            + [self.lvm_node]



#
# Copyright 2017 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# Scan the element directory, looks for element  dependencies and
# writes them out when the directive
#   .. element_deps::
# is used.
# This was developed only for internal use and must be called
# from the source top directory.

from diskimage_builder.element_dependencies import _find_all_elements
from diskimage_builder.paths import get_path

from docutils.parsers.rst import Directive
import os

all_elements = _find_all_elements(get_path("elements"))


def make_dep_list(title, deps):
    lines = []
    lines.append(title)
    lines.append("+" * len(title))
    for dep in sorted(deps):
        lines.append("* :doc:`../%s/README`" % dep)
    lines.append('')  # careful to end with a blank line
    return lines


class ElementDepsDirective(Directive):

    # this enables content in the directive
    has_content = True

    def run(self):
        source = self.state_machine.input_lines.source(
            self.lineno - self.state_machine.input_offset - 1)
        # Extract the element from the source attribute of the document
        element_name = os.path.basename(os.path.dirname(
            self.state_machine.document.attributes['source']))

        lines = ["Element Dependencies",
                 "--------------------"]

        # This should not fail -- sphinx would be finding an element
        # that dib doesn't know about?
        element = all_elements[element_name]
        if element.depends:
            lines.extend(make_dep_list("Uses", element.depends))
        if element.r_depends:
            lines.extend(make_dep_list("Used by", element.r_depends))

        self.state_machine.insert_input(lines, source)

        return []


def setup(app):
    app.add_directive('element_deps', ElementDepsDirective)



# Copyright 2016 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import diskimage_builder.block_device.tests.test_base as tb

from diskimage_builder.block_device.utils import parse_abs_size_spec
from diskimage_builder.block_device.utils import parse_rel_size_spec


logger = logging.getLogger(__name__)


class TestBlockDeviceUtils(tb.TestBase):
    """Tests for the utils.py

    This tests mostly the error and failure cases - because the good
    cases are tested implicitly with the higher level unit tests.
    """

    def test_parse_rel_size_with_abs(self):
        """Calls parse_rel_size_spec with an absolute number"""

        is_rel, size = parse_rel_size_spec("154MiB", 0)
        self.assertFalse(is_rel)
        self.assertEqual(154 * 1024 * 1024, size)

    def test_parse_abs_size_without_spec(self):
        """Call parse_abs_size_spec without spec"""

        size = parse_abs_size_spec("198")
        self.assertEqual(198, size)

    def test_invalid_unit_spec(self):
        """Call parse_abs_size_spec with invalid unit spec"""

        self.assertRaises(RuntimeError, parse_abs_size_spec, "747InVaLiDUnIt")

    def test_broken_unit_spec(self):
        """Call parse_abs_size_spec with a completely broken unit spec"""

        self.assertRaises(RuntimeError, parse_abs_size_spec, "_+!HuHi+-=")

    def test_parse_size_spec(self):
        map(lambda tspec:
            self.assertEqual(parse_abs_size_spec(tspec[0]), tspec[1]),
            [["20TiB", 20 * 1024**4],
             ["1024KiB", 1024 * 1024],
             ["1.2TB", 1.2 * 1000**4],
             ["2.4T", 2.4 * 1000**4],
             ["512B", 512],
             ["364", 364]])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

import diskimage_builder.block_device.tests.test_config as tc

from diskimage_builder.block_device.config import create_graph
from diskimage_builder.block_device.exception import \
    BlockDeviceSetupException


logger = logging.getLogger(__name__)


class TestMkfs(tc.TestGraphGeneration):

    def test_duplicate_labels(self):

        config = self.load_config_file('duplicate_fs_labels.yaml')

        self.assertRaisesRegex(BlockDeviceSetupException,
                               "used more than once",
                               create_graph, config,
                               self.fake_default_config, {})

    def test_too_long_labels(self):

        config = self.load_config_file('too_long_fs_label.yaml')

        self.assertRaisesRegex(BlockDeviceSetupException,
                               "too long for filesystem",
                               create_graph, config,
                               self.fake_default_config, {})



# Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import imp
import mock
import os
from oslotest import base

module_path = (os.path.dirname(os.path.realpath(__file__)) +
               '/../static/usr/local/sbin/growvols')
growvols = imp.load_source('growvols', module_path)

# output of lsblk -Po kname,pkname,name,label,type,fstype,mountpoint
LSBLK = """KNAME="sda" PKNAME="" NAME="sda" LABEL="" TYPE="disk" FSTYPE="" MOUNTPOINT=""
KNAME="sda1" PKNAME="sda" NAME="sda1" LABEL="MKFS_ESP" TYPE="part" FSTYPE="vfat" MOUNTPOINT="/boot/efi"
KNAME="sda2" PKNAME="sda" NAME="sda2" LABEL="" TYPE="part" FSTYPE="" MOUNTPOINT=""
KNAME="sda3" PKNAME="sda" NAME="sda3" LABEL="" TYPE="part" FSTYPE="LVM2_member" MOUNTPOINT=""
KNAME="sda4" PKNAME="sda" NAME="sda4" LABEL="config-2" TYPE="part" FSTYPE="iso9660" MOUNTPOINT=""
KNAME="dm-0" PKNAME="sda3" NAME="vg-lv_root" LABEL="img-rootfs" TYPE="lvm" FSTYPE="xfs" MOUNTPOINT="/"
KNAME="dm-1" PKNAME="sda3" NAME="vg-lv_tmp" LABEL="fs_tmp" TYPE="lvm" FSTYPE="xfs" MOUNTPOINT="/tmp"
KNAME="dm-2" PKNAME="sda3" NAME="vg-lv_var" LABEL="fs_var" TYPE="lvm" FSTYPE="xfs" MOUNTPOINT="/var"
KNAME="dm-3" PKNAME="sda3" NAME="vg-lv_home" LABEL="fs_home" TYPE="lvm" FSTYPE="xfs" MOUNTPOINT="/home"
""" # noqa

DEVICES = [{
    "FSTYPE": "",
    "KNAME": "sda",
    "LABEL": "",
    "MOUNTPOINT": "",
    "NAME": "sda",
    "PKNAME": "",
    "TYPE": "disk",
}, {
    "FSTYPE": "vfat",
    "KNAME": "sda1",
    "LABEL": "MKFS_ESP",
    "MOUNTPOINT": "/boot/efi",
    "NAME": "sda1",
    "PKNAME": "sda",
    "TYPE": "part",
}, {
    "FSTYPE": "",
    "KNAME": "sda2",
    "LABEL": "",
    "MOUNTPOINT": "",
    "NAME": "sda2",
    "PKNAME": "sda",
    "TYPE": "part",
}, {
    "FSTYPE": "LVM2_member",
    "KNAME": "sda3",
    "LABEL": "",
    "MOUNTPOINT": "",
    "NAME": "sda3",
    "PKNAME": "sda",
    "TYPE": "part",
}, {
    "FSTYPE": "iso9660",
    "KNAME": "sda4",
    "LABEL": "config-2",
    "MOUNTPOINT": "",
    "NAME": "sda4",
    "PKNAME": "sda",
    "TYPE": "part",
}, {
    "FSTYPE": "xfs",
    "KNAME": "dm-0",
    "LABEL": "img-rootfs",
    "MOUNTPOINT": "/",
    "NAME": "vg-lv_root",
    "PKNAME": "sda3",
    "TYPE": "lvm",
}, {
    "FSTYPE": "xfs",
    "KNAME": "dm-1",
    "LABEL": "fs_tmp",
    "MOUNTPOINT": "/tmp",
    "NAME": "vg-lv_tmp",
    "PKNAME": "sda3",
    "TYPE": "lvm",
}, {
    "FSTYPE": "xfs",
    "KNAME": "dm-2",
    "LABEL": "fs_var",
    "MOUNTPOINT": "/var",
    "NAME": "vg-lv_var",
    "PKNAME": "sda3",
    "TYPE": "lvm",
}, {
    "FSTYPE": "xfs",
    "KNAME": "dm-3",
    "LABEL": "fs_home",
    "MOUNTPOINT": "/home",
    "NAME": "vg-lv_home",
    "PKNAME": "sda3",
    "TYPE": "lvm",
}]

# output of sgdisk --first-aligned-in-largest --end-of-largest /dev/sda
SECTOR_START = 79267840
SECTOR_END = 488265727
SGDISK_LARGEST = "%s\n%s\n" % (SECTOR_START, SECTOR_END)
SGDISK_V = """
Problem: The secondary header's self-pointer indicates that it doesn't reside
at the end of the disk. If you've added a disk to a RAID array, use the 'e'
option on the experts' menu to adjust the secondary header's and partition
table's locations.

Identified 1 problems!"""

# output of vgs --noheadings --options vg_name
VGS = "  vg\n"

# output of lvs --noheadings --options lv_name,lv_dm_path,lv_attr,pool_lv
LVS = '''
  lv_audit    /dev/mapper/vg-lv_audit    Vwi-aotz--
  lv_home     /dev/mapper/vg-lv_home     Vwi-aotz--
  lv_log      /dev/mapper/vg-lv_log      Vwi-aotz--
  lv_root     /dev/mapper/vg-lv_root     Vwi-aotz--
  lv_srv      /dev/mapper/vg-lv_srv      Vwi-aotz--
  lv_tmp      /dev/mapper/vg-lv_tmp      Vwi-aotz--
  lv_var      /dev/mapper/vg-lv_var      Vwi-aotz--
'''

LVS_THIN = '''
  lv_audit    /dev/mapper/vg-lv_audit    Vwi-aotz-- lv_thinpool
  lv_home     /dev/mapper/vg-lv_home     Vwi-aotz-- lv_thinpool
  lv_log      /dev/mapper/vg-lv_log      Vwi-aotz-- lv_thinpool
  lv_root     /dev/mapper/vg-lv_root     Vwi-aotz-- lv_thinpool
  lv_srv      /dev/mapper/vg-lv_srv      Vwi-aotz-- lv_thinpool
  lv_thinpool /dev/mapper/vg-lv_thinpool twi-aotz--
  lv_tmp      /dev/mapper/vg-lv_tmp      Vwi-aotz-- lv_thinpool
  lv_var      /dev/mapper/vg-lv_var      Vwi-aotz-- lv_thinpool
'''


class TestGrowvols(base.BaseTestCase):

    def test_printable_cmd(self):
        self.assertEqual(
            "foo --thing 'bar baz'",
            growvols.printable_cmd(['foo', '--thing', "bar baz"])
        )

    def test_convert_bytes(self):
        self.assertEqual('100B', growvols.convert_bytes(100))
        self.assertEqual('1000B', growvols.convert_bytes(1000))
        self.assertEqual('1MiB', growvols.convert_bytes(2000000))
        self.assertEqual('2GiB', growvols.convert_bytes(3000000000))
        self.assertEqual('3725GiB', growvols.convert_bytes(4000000000000))

    @mock.patch('subprocess.Popen')
    def test_execute(self, mock_popen):
        mock_process = mock.Mock()
        mock_process.returncode = 0
        mock_process.communicate.return_value = ('did the thing', '')
        mock_popen.return_value = mock_process

        result = growvols.execute(['do', 'the', 'thing'])
        self.assertEqual('did the thing', result)

        mock_process.returncode = 1
        mock_process.communicate.return_value = ('', 'ouch')

        e = self.assertRaises(Exception, growvols.execute,
                              ['do', 'the', 'thing'])
        self.assertIn('ouch', str(e))

    def test_parse_shell_vars(self):
        devices = list(growvols.parse_shell_vars(LSBLK))
        self.assertEqual(DEVICES, devices)

    def test_find_device(self):
        sda = {
            "FSTYPE": "",
            "KNAME": "sda",
            "LABEL": "",
            "MOUNTPOINT": "",
            "NAME": "sda",
            "PKNAME": "",
            "TYPE": "disk",
        }
        fs_home = {
            "FSTYPE": "xfs",
            "KNAME": "dm-3",
            "LABEL": "fs_home",
            "MOUNTPOINT": "/home",
            "NAME": "vg-lv_home",
            "PKNAME": "sda3",
            "TYPE": "lvm",
        }
        self.assertEqual(
            sda, growvols.find_device(DEVICES, 'NAME', 'sda'))
        self.assertEqual(
            fs_home,
            growvols.find_device(
                DEVICES, ['KNAME', 'NAME'], 'vg-lv_home'))
        self.assertEqual(
            fs_home,
            growvols.find_device(
                DEVICES, ['KNAME', 'NAME'], 'dm-3'))
        self.assertIsNone(
            growvols.find_device(
                DEVICES, ['KNAME', 'NAME'], 'asdf'))

    def test_find_disk(self):
        devices = list(growvols.parse_shell_vars(LSBLK))
        opts = mock.Mock()
        opts.device = None
        sda = growvols.find_device(devices, 'NAME', 'sda')

        # discover via MOUNTPOINT /
        self.assertEqual(sda, growvols.find_disk(opts, devices))

        # fetch sda
        opts.device = 'sda'
        self.assertEqual(sda, growvols.find_disk(opts, devices))

        # delete sda3, so can't find relationship
        # from MOUNTPOINT / to sda
        opts.device = None
        devices = [d for d in devices if d['NAME'] != 'sda3']
        e = self.assertRaises(Exception, growvols.find_disk, opts, devices)
        self.assertEqual('Could not detect disk device', str(e))

        # no sdb
        opts.device = 'sdb'
        e = self.assertRaises(Exception, growvols.find_disk, opts, devices)
        self.assertEqual('Could not find specified --device: sdb', str(e))

        # sda is not TYPE disk
        sda['TYPE'] = 'dissed'
        opts.device = 'sda'
        e = self.assertRaises(Exception, growvols.find_disk, opts, devices)
        self.assertEqual('Expected a device with TYPE="disk", got: dissed',
                         str(e))

    @mock.patch('growvols.execute')
    def test_find_space(self, mock_execute):
        mock_execute.return_value = SGDISK_LARGEST
        sector_start, sector_end, size_sectors = growvols.find_space(
            'sda')
        self.assertEqual(SECTOR_START, sector_start)
        self.assertEqual(SECTOR_END, sector_end)
        self.assertEqual(SECTOR_END - SECTOR_START, size_sectors)
        mock_execute.assert_called_once_with([
            'sgdisk',
            '--first-aligned-in-largest',
            '--end-of-largest',
            '/dev/sda'])

    @mock.patch('growvols.execute')
    def test_find_devices(self, mock_execute):
        mock_execute.return_value = LSBLK
        self.assertEqual(DEVICES, growvols.find_devices())
        mock_execute.assert_called_once_with([
            'lsblk',
            '-Po',
            'kname,pkname,name,label,type,fstype,mountpoint'])

    @mock.patch('growvols.execute')
    def test_find_group(self, mock_execute):
        mock_execute.return_value = VGS
        opts = mock.Mock()
        opts.group = None
        self.assertEqual('vg', growvols.find_group(opts))
        mock_execute.assert_called_once_with([
            'vgs', '--noheadings', '--options', 'vg_name'])

        # no volume groups
        mock_execute.return_value = "\n"
        e = self.assertRaises(Exception, growvols.find_group, opts)
        self.assertEqual('No volume groups found', str(e))

        # multiple volume groups
        mock_execute.return_value = "  vg\nvg2\nvg3"
        e = self.assertRaises(Exception, growvols.find_group, opts)
        self.assertEqual('More than one volume group, specify one to '
                         'use with --group: vg, vg2, vg3', str(e))

        # multiple volume groups with explicit group argument
        opts.group = 'vg'
        self.assertEqual('vg', growvols.find_group(opts))

        # no such group
        opts.group = 'novg'
        e = self.assertRaises(Exception, growvols.find_group, opts)
        self.assertEqual('Could not find specified --group: novg', str(e))

    def test_find_next_partnum(self):
        self.assertEqual(5, growvols.find_next_partnum(DEVICES, 'sda'))
        self.assertEqual(1, growvols.find_next_partnum(DEVICES, 'sdb'))

    def test_find_next_device_name(self):
        devices = list(growvols.parse_shell_vars(LSBLK))

        # Use SATA etc device naming
        self.assertEqual(
            'sda5',
            growvols.find_next_device_name(devices, 'sda', 5))

        # No partitions
        e = self.assertRaises(Exception, growvols.find_next_device_name,
                              devices, 'sdb', 1)
        self.assertEqual(
            'Could not find partition naming scheme for sdb', str(e))

        # Use NVMe device naming
        for i in (1, 2, 3, 4):
            d = growvols.find_device(devices, 'KNAME', 'sda%s' % i)
            d['KNAME'] = 'nvme0p%s' % i
        self.assertEqual(
            'nvme0p5',
            growvols.find_next_device_name(devices, 'nvme0', 5))

    def test_amount_unit_to_extent(self):
        one_m = growvols.UNIT_BYTES['MiB']
        four_m = one_m * 4
        one_g = growvols.UNIT_BYTES['GiB']
        ten_g = one_g * 10
        forty_g = ten_g * 4
        fidy_g = ten_g * 5

        # invalid amounts
        self.assertRaises(Exception, growvols.amount_unit_to_extent,
                          '100', one_g, one_g)
        self.assertRaises(Exception, growvols.amount_unit_to_extent,
                          '100B', one_g, one_g)
        self.assertRaises(Exception, growvols.amount_unit_to_extent,
                          '100%%', one_g, one_g)
        self.assertRaises(Exception, growvols.amount_unit_to_extent,
                          '100TiB', one_g, one_g)
        self.assertRaises(Exception, growvols.amount_unit_to_extent,
                          'i100MB', one_g, one_g)

        # GiB amount
        self.assertEqual(
            (ten_g, forty_g),
            growvols.amount_unit_to_extent('10GiB', fidy_g, fidy_g)
        )

        # percentage amount
        self.assertEqual(
            (ten_g, forty_g),
            growvols.amount_unit_to_extent('20%', fidy_g, fidy_g)
        )

        # not enough space left
        self.assertEqual(
            (0, one_m),
            growvols.amount_unit_to_extent('20%', fidy_g, one_m)
        )

        # exactly one extent
        self.assertEqual(
            (four_m, fidy_g - four_m),
            growvols.amount_unit_to_extent('4MiB', fidy_g, fidy_g)
        )

        # under one extent is zero
        self.assertEqual(
            (0, fidy_g),
            growvols.amount_unit_to_extent('3MiB', fidy_g, fidy_g)
        )

        # just over one extent is one extent
        self.assertEqual(
            (four_m, fidy_g - four_m),
            growvols.amount_unit_to_extent('5MiB', fidy_g, fidy_g)
        )

    def test_find_grow_vols(self):
        one_g = growvols.UNIT_BYTES['GiB']
        ten_g = one_g * 10
        fidy_g = ten_g * 5

        opts = mock.Mock()

        # buy default, assign all to /
        opts.grow_vols = ['']
        self.assertEqual(
            {'/dev/mapper/vg-lv_root': fidy_g},
            growvols.find_grow_vols(opts, DEVICES, 'vg', fidy_g)
        )

        # assign to /home, /var, remainder to /
        opts.grow_vols = ['/home=20%', 'fs_var=40%']
        self.assertEqual(
            {
                '/dev/mapper/vg-lv_home': ten_g,
                '/dev/mapper/vg-lv_var': ten_g * 2,
                '/dev/mapper/vg-lv_root': ten_g * 2
            },
            growvols.find_grow_vols(opts, DEVICES, 'vg', fidy_g)
        )

        # assign to /home, /var, /tmp by amount
        opts.grow_vols = ['/home=19GiB', 'fs_var=30GiB', '/tmp=1GiB']
        self.assertEqual(
            {
                '/dev/mapper/vg-lv_home': one_g * 19,
                '/dev/mapper/vg-lv_var': one_g * 30,
                '/dev/mapper/vg-lv_tmp': one_g
            },
            growvols.find_grow_vols(opts, DEVICES, 'vg', fidy_g)
        )

    @mock.patch('builtins.open', autospec=True)
    def test_find_sector_size(self, mock_open):
        mock_open.return_value.__enter__ = lambda s: s
        mock_open.return_value.__exit__ = mock.Mock()
        read_mock = mock_open.return_value.read
        read_mock.side_effect = ['512']

        # disk sdx exists
        self.assertEqual(512, growvols.find_sector_size('sdx'))

        # disk sdx doesn't exist
        mock_open.side_effect = FileNotFoundError
        self.assertRaises(FileNotFoundError, growvols.find_sector_size, 'sdx')

    @mock.patch('growvols.execute')
    def test_find_thin_pool(self, mock_execute):
        # No thin pool
        mock_execute.return_value = LVS
        self.assertIsNone(growvols.find_thin_pool(DEVICES, 'vg'))
        mock_execute.assert_called_once_with([
            'lvs', '--noheadings', '--options',
            'lv_name,lv_dm_path,lv_attr,pool_lv'])

        # One thin pool, all volumes use it
        mock_execute.return_value = LVS_THIN
        self.assertEqual('/dev/mapper/vg-lv_thinpool',
                         growvols.find_thin_pool(DEVICES, 'vg'))

        # One pool, not used by all volumes
        mock_execute.return_value = '''
  lv_thinpool /dev/mapper/vg-lv_thinpool twi-aotz--
  lv_home     /dev/mapper/vg-lv_home     Vwi-aotz--
  lv_root     /dev/mapper/vg-lv_root     Vwi-aotz-- lv_thinpool'''
        e = self.assertRaises(Exception, growvols.find_thin_pool,
                              DEVICES, 'vg')
        self.assertEqual('All volumes need to be in pool lv_thinpool. '
                         'lv_home is in pool None', str(e))

        # Two pools, volumes use both
        mock_execute.return_value = '''
  lv_thin1 /dev/mapper/vg-lv_thin1 twi-aotz--
  lv_thin2 /dev/mapper/vg-lv_thin2 twi-aotz--
  lv_home  /dev/mapper/vg-lv_home  Vwi-aotz-- lv_thin2
  lv_root  /dev/mapper/vg-lv_root  Vwi-aotz-- lv_thin1'''
        e = self.assertRaises(Exception, growvols.find_thin_pool,
                              DEVICES, 'vg')
        self.assertEqual('All volumes need to be in pool lv_thin1. '
                         'lv_home is in pool lv_thin2', str(e))

    @mock.patch('growvols.find_sector_size')
    @mock.patch('growvols.execute')
    def test_main(self, mock_execute, mock_sector_size):
        mock_sector_size.return_value = 512

        # noop, only discover block device info
        mock_execute.side_effect = [
            LSBLK,
            SGDISK_V,
            '',
            '',
            SGDISK_LARGEST,
            VGS,
            LVS,
        ]
        growvols.main(['growvols', '--noop'])
        mock_execute.assert_has_calls([
            mock.call(['lsblk', '-Po',
                       'kname,pkname,name,label,type,fstype,mountpoint']),
            mock.call(['sgdisk', '-v', '/dev/sda']),
            mock.call(['sgdisk', '-e', '/dev/sda']),
            mock.call(['partprobe']),
            mock.call(['sgdisk', '--first-aligned-in-largest',
                       '--end-of-largest', '/dev/sda']),
            mock.call(['vgs', '--noheadings', '--options', 'vg_name']),
            mock.call(['lvs', '--noheadings', '--options',
                       'lv_name,lv_dm_path,lv_attr,pool_lv'])
        ])

        # no arguments, assign all to /
        mock_execute.reset_mock()
        mock_execute.side_effect = [
            LSBLK,
            '',
            SGDISK_LARGEST,
            VGS,
            LVS,
            '', '', '', '', '', ''
        ]
        growvols.main(['growvols', '--yes'])
        mock_execute.assert_has_calls([
            mock.call(['lsblk', '-Po',
                       'kname,pkname,name,label,type,fstype,mountpoint']),
            mock.call(['sgdisk', '-v', '/dev/sda']),
            mock.call(['sgdisk', '--first-aligned-in-largest',
                       '--end-of-largest', '/dev/sda']),
            mock.call(['vgs', '--noheadings', '--options', 'vg_name']),
            mock.call(['lvs', '--noheadings', '--options',
                       'lv_name,lv_dm_path,lv_attr,pool_lv']),
            mock.call(['sgdisk', '--new=5:79267840:488265727',
                       '--change-name=5:growvols', '/dev/sda']),
            mock.call(['partprobe']),
            mock.call(['pvcreate', '/dev/sda5']),
            mock.call(['vgextend', 'vg', '/dev/sda5']),
            mock.call(['lvextend', '--size', '+209404821504B',
                       '/dev/mapper/vg-lv_root', '/dev/sda5']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_root'])
        ])

        # assign to /home, /var, remainder to /
        mock_execute.reset_mock()
        mock_execute.side_effect = [
            LSBLK,
            '',
            SGDISK_LARGEST,
            VGS,
            LVS,
            '', '', '', '', '', '', '', '', '', ''
        ]
        growvols.main(['growvols', '--yes', '--group', 'vg',
                       '/home=20%', 'fs_var=40%'])
        mock_execute.assert_has_calls([
            mock.call(['lsblk', '-Po',
                       'kname,pkname,name,label,type,fstype,mountpoint']),
            mock.call(['sgdisk', '-v', '/dev/sda']),
            mock.call(['sgdisk', '--first-aligned-in-largest',
                       '--end-of-largest', '/dev/sda']),
            mock.call(['vgs', '--noheadings', '--options', 'vg_name']),
            mock.call(['lvs', '--noheadings', '--options',
                       'lv_name,lv_dm_path,lv_attr,pool_lv']),
            mock.call(['sgdisk', '--new=5:79267840:488265727',
                       '--change-name=5:growvols', '/dev/sda']),
            mock.call(['partprobe']),
            mock.call(['pvcreate', '/dev/sda5']),
            mock.call(['vgextend', 'vg', '/dev/sda5']),
            mock.call(['lvextend', '--size', '+41880125440B',
                       '/dev/mapper/vg-lv_home', '/dev/sda5']),
            mock.call(['lvextend', '--size', '+83760250880B',
                       '/dev/mapper/vg-lv_var', '/dev/sda5']),
            mock.call(['lvextend', '--size', '+83764445184B',
                       '/dev/mapper/vg-lv_root', '/dev/sda5']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_home']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_var']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_root']),
        ])

        # no space to grow, failed
        sector_start = 79267840
        sector_end = sector_start + 1024
        sgdisk_largest = "%s\n%s\n" % (sector_start, sector_end)
        mock_execute.side_effect = [
            LSBLK,
            '',
            sgdisk_largest,
            VGS,
            LVS,
        ]
        self.assertEqual(
            2,
            growvols.main(['growvols', '--exit-on-no-grow'])
        )

        # no space to grow, success
        mock_execute.side_effect = [
            LSBLK,
            '',
            sgdisk_largest,
            VGS,
            LVS,
        ]
        self.assertEqual(
            0,
            growvols.main(['growvols'])
        )

    @mock.patch('growvols.find_sector_size')
    @mock.patch('growvols.execute')
    def test_main_thin_provision(self, mock_execute, mock_sector_size):
        mock_sector_size.return_value = 512

        # assign to /home, /var, remainder to /
        mock_execute.reset_mock()
        mock_execute.side_effect = [
            LSBLK,
            '',
            SGDISK_LARGEST,
            VGS,
            LVS_THIN,
            '', '', '', '', '', '', '', '', '', '', '', ''
        ]
        growvols.main(['growvols', '--yes', '--group', 'vg',
                       '/home=20%', 'fs_var=40%'])
        mock_execute.assert_has_calls([
            mock.call(['lsblk', '-Po',
                       'kname,pkname,name,label,type,fstype,mountpoint']),
            mock.call(['sgdisk', '-v', '/dev/sda']),
            mock.call(['sgdisk', '--first-aligned-in-largest',
                       '--end-of-largest', '/dev/sda']),
            mock.call(['vgs', '--noheadings', '--options', 'vg_name']),
            mock.call(['lvs', '--noheadings', '--options',
                       'lv_name,lv_dm_path,lv_attr,pool_lv']),
            mock.call(['sgdisk', '--new=5:79267840:488265727',
                       '--change-name=5:growvols', '/dev/sda']),
            mock.call(['partprobe']),
            mock.call(['pvcreate', '/dev/sda5']),
            mock.call(['vgextend', 'vg', '/dev/sda5']),
            mock.call(['lvextend', '--poolmetadatasize', '+1073741824B',
                       '/dev/mapper/vg-lv_thinpool', '/dev/sda5']),
            mock.call(['lvextend', '-L+208326885376B',
                       '/dev/mapper/vg-lv_thinpool', '/dev/sda5']),
            mock.call(['lvextend', '--size', '+41662021632B',
                       '/dev/mapper/vg-lv_home']),
            mock.call(['lvextend', '--size', '+83328237568B',
                       '/dev/mapper/vg-lv_var']),
            mock.call(['lvextend', '--size', '+83336626176B',
                       '/dev/mapper/vg-lv_root']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_home']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_var']),
            mock.call(['xfs_growfs', '/dev/mapper/vg-lv_root']),
        ])



# Copyright 2014 Red Hat, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# TODO(bnemec): This whole file is a copy-paste of the one in
# tripleo-image-elements.  That duplication needs to be eliminated.

import os
import subprocess
import sys
import tempfile

from oslotest import base


class ScriptTestBase(base.BaseTestCase):
    def setUp(self):
        super(ScriptTestBase, self).setUp()
        self.tmpdir = tempfile.mkdtemp()
        self.env = os.environ.copy()
        self.env['PATH'] = self.tmpdir + ':' + self.env['PATH']

    def _stub_script(self, name, contents):
        filename = os.path.join(self.tmpdir, name)
        with open(filename, 'w') as f:
            f.write('#!/bin/bash\n')
            f.write(contents)
            f.write('\n')
        os.chmod(filename, 0o700)

    def _run_command(self, cmd):
        try:
            # check_output doesn't exist in Python < 2.7
            if sys.hexversion < 0x02070000:
                runner = subprocess.check_call
            else:
                runner = subprocess.check_output
            return runner(cmd,
                          stderr=subprocess.STDOUT,
                          env=self.env)
        # NOTE(bnemec): Handle the exception so we can extract as much
        # information as possible.
        except subprocess.CalledProcessError as e:
            # The check_call exception won't have any data in e.output
            if sys.hexversion < 0x02070000:
                self.fail(e)
            else:
                self.fail(e.output)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import logging
import mock
import os

import diskimage_builder.block_device.tests.test_config as tc

from diskimage_builder.block_device.config import config_tree_to_graph
from diskimage_builder.block_device.config import create_graph
from diskimage_builder.block_device.level2.mkfs import FilesystemNode
from diskimage_builder.block_device.level3.mount import cmp_mount_order
from diskimage_builder.block_device.level3.mount import MountPointNode
from diskimage_builder.block_device.tests.test_base import TestBase

logger = logging.getLogger(__name__)


class TestMountComparator(TestBase):

    def test_mount_comparator(self):
        # This tests cmp_mount_order to ensure it sorts in the
        # expected order.  The comparator takes a tuple of
        # (mount_point, node_name) but we can ignore the name
        partitions = [
            ('/var/log', 'fake_log'),
            ('/boot', 'fake_boot'),
            ('/', 'fake_name'),
            ('/var', 'fake_name')]
        partitions.sort(key=functools.cmp_to_key(cmp_mount_order))

        res = list(x[0] for x in partitions)

        # "/" must be first
        self.assertEqual(res[0], '/')

        # /var before /var/log
        var_pos = res.index('/var')
        var_log_pos = res.index('/var/log')
        self.assertGreater(var_log_pos, var_pos)


class TestMountOrder(tc.TestGraphGeneration):

    def _exec_sudo_log(*args, **kwargs):
        # Used as a side-effect from exec_sudo mocking so we can see
        # the call in-place in logs
        logger.debug("exec_sudo: %s", " ".join(args[0]))

    @mock.patch('diskimage_builder.block_device.level3.mount.exec_sudo',
                side_effect=_exec_sudo_log)
    @mock.patch('diskimage_builder.block_device.level2.mkfs.exec_sudo',
                side_effect=_exec_sudo_log)
    def test_mfks_and_mount_order(self, mock_exec_sudo_mkfs,
                                  mock_exec_sudo_mount):
        # XXX: better mocking for the os.path.exists calls to avoid
        # failing if this exists.
        self.assertFalse(os.path.exists('/fake/'))

        # This is probably in order after graph creation, so ensure it
        # remains stable.  We test the mount and umount call sequences
        config = self.load_config_file('multiple_partitions_graph.yaml')
        state = {}
        graph, call_order = create_graph(config, self.fake_default_config,
                                         state)

        # Mocked block device state
        state['blockdev'] = {}
        state['blockdev']['root'] = {'device': '/dev/loopXp1/root'}
        state['blockdev']['var'] = {'device': '/dev/loopXp2/var'}
        state['blockdev']['var_log'] = {'device': '/dev/loopXp3/var_log'}
        state['blockdev']['swap'] = {'device': '/dev/loopXp4/swap'}

        for node in call_order:
            if isinstance(node, (FilesystemNode, MountPointNode)):
                node.create()
        for node in reversed(call_order):
            if isinstance(node, (FilesystemNode, MountPointNode)):
                node.umount()

        # ensure that partitions were mounted in order root->var->var/log
        self.assertListEqual(state['mount_order'],
                             ['/', '/var', '/var/log', 'none'])

        # fs creation sequence (note we don't care about order of this
        # as they're all independent)
        cmd_sequence = [
            mock.call(['mkfs', '-t', 'xfs', '-L', 'mkfs_root',
                       '-m', 'uuid=root-uuid-1234',
                       '-q', '/dev/loopXp1/root']),
            mock.call(['mkfs', '-t', 'xfs', '-L', 'mkfs_var',
                       '-m', 'uuid=var-uuid-1234',
                       '-q', '/dev/loopXp2/var']),
            mock.call(['mkfs', '-t', 'vfat', '-n', 'VARLOG',
                       '/dev/loopXp3/var_log']),
            mock.call(['mkswap', '-L', 'mkfs_swap', '-U', 'swap-uuid-1234',
                       '/dev/loopXp4/swap'])
        ]
        self.assertEqual(mock_exec_sudo_mkfs.call_count, len(cmd_sequence))
        mock_exec_sudo_mkfs.assert_has_calls(cmd_sequence, any_order=True)

        # Check mount sequence
        cmd_sequence = [
            # mount sequence
            mock.call(['mkdir', '-p', '/fake/']),
            mock.call(['mount', '/dev/loopXp1/root', '/fake/']),
            mock.call(['mkdir', '-p', '/fake/var']),
            mock.call(['mount', '/dev/loopXp2/var', '/fake/var']),
            mock.call(['mkdir', '-p', '/fake/var/log']),
            mock.call(['mount', '/dev/loopXp3/var_log', '/fake/var/log']),
            # umount sequence
            mock.call(['sync']),
            # note /fake/var/log is a vfs partition to make sure
            # we don't try to fstrim it
            mock.call(['umount', '/fake/var/log']),
            mock.call(['sync']),
            mock.call(['fstrim', '--verbose', '/fake/var']),
            mock.call(['umount', '/fake/var']),
            mock.call(['sync']),
            mock.call(['fstrim', '--verbose', '/fake/']),
            mock.call(['umount', '/fake/'])
        ]
        self.assertListEqual(mock_exec_sudo_mount.call_args_list, cmd_sequence)

    @mock.patch('diskimage_builder.block_device.level3.mount.exec_sudo',
                side_effect=_exec_sudo_log)
    def test_mount_order_unsorted(self, mock_exec_sudo):
        # As above, but this is out of order and gets sorted
        # so that root is mounted first (and skips the mkfs testing).
        config = self.load_config_file('lvm_tree_partition_ordering.yaml')
        parsed_graph = config_tree_to_graph(config)
        state = {}

        graph, call_order = create_graph(parsed_graph,
                                         self.fake_default_config,
                                         state)
        state['filesys'] = {
            'mkfs_root': {
                'device': '/dev/loopXp1',
                'fstype': 'xfs'
            },
            'mkfs_var': {
                'device': '/dev/loopXp2',
                'fstype': 'xfs',
            },
            'mkfs_boot': {
                'device': '/dev/loopXp3',
                'fstype': 'vfat',
            },
            'mkfs_swap': {
                'device': '/dev/loopXp4',
                'fstype': 'swap',
            },
        }

        for node in call_order:
            if isinstance(node, MountPointNode):
                node.create()
        for node in reversed(call_order):
            if isinstance(node, MountPointNode):
                node.umount()

        # ensure that partitions are mounted in order / -> /boot -> /var
        # swap not mounted in runtime, but should be in 'mount_order' list
        self.assertListEqual(state['mount_order'],
                             ['/', '/boot', '/var', 'none'])

        cmd_sequence = [
            # mount sequence
            mock.call(['mkdir', '-p', '/fake/']),
            mock.call(['mount', '/dev/loopXp1', '/fake/']),
            mock.call(['mkdir', '-p', '/fake/boot']),
            mock.call(['mount', '/dev/loopXp3', '/fake/boot']),
            mock.call(['mkdir', '-p', '/fake/var']),
            mock.call(['mount', '/dev/loopXp2', '/fake/var']),
            # umount sequence
            mock.call(['sync']),
            mock.call(['fstrim', '--verbose', '/fake/var']),
            mock.call(['umount', '/fake/var']),
            mock.call(['sync']),
            # no trim on vfat /fake/boot
            mock.call(['umount', '/fake/boot']),
            mock.call(['sync']),
            mock.call(['fstrim', '--verbose', '/fake/']),
            mock.call(['umount', '/fake/'])
        ]
        self.assertListEqual(mock_exec_sudo.call_args_list, cmd_sequence)



# Copyright 2016 Andreas Florath (andreas@florath.net)
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

from diskimage_builder.block_device.exception import \
    BlockDeviceSetupException
from diskimage_builder.block_device.plugin import NodeBase
from diskimage_builder.block_device.plugin import PluginBase
from diskimage_builder.block_device.utils import exec_sudo
from diskimage_builder.block_device.utils import parse_abs_size_spec


logger = logging.getLogger(__name__)


def image_create(filename, size):
    logger.info("Create image file [%s]", filename)
    with open(filename, "w") as fd:
        fd.seek(size - 1)
        fd.write("\0")


def image_delete(filename):
    logger.info("Remove image file [%s]", filename)
    os.remove(filename)


def loopdev_attach(filename):
    logger.info("loopdev attach")
    logger.debug("Calling [sudo losetup --show -f %s]", filename)
    block_device = exec_sudo(["losetup", "--show", "-f", filename])
    # [:-1]: Cut of the newline
    block_device = block_device[:-1]
    logger.info("New block device [%s]", block_device)
    return block_device


def loopdev_detach(loopdev):
    logger.info("loopdev detach")
    # loopback dev may be tied up a bit by udev events triggered
    # by partition events
    for try_cnt in range(10, 1, -1):
        try:
            exec_sudo(["losetup", "-d", loopdev])
            return
        except BlockDeviceSetupException as e:
            # Do not raise an error - maybe other cleanup methods
            # can at least do some more work.
            logger.error("loopdev detach failed (%s)", e.returncode)

    logger.debug("Gave up trying to detach [%s]", loopdev)
    return 1


class LocalLoopNode(NodeBase):
    """Level0: Local loop image device handling.

    This class handles local loop devices that can be used
    for VM image installation.
    """
    def __init__(self, config, default_config, state):
        logger.debug("Creating LocalLoop object; config [%s] "
                     "default_config [%s]", config, default_config)
        super(LocalLoopNode, self).__init__(config['name'], state)
        if 'size' in config:
            self.size = parse_abs_size_spec(config['size'])
            logger.debug("Image size [%s]", self.size)
        else:
            self.size = parse_abs_size_spec(default_config['image-size'])
            logger.debug("Using default image size [%s]", self.size)
        if 'directory' in config:
            self.image_dir = config['directory']
        else:
            self.image_dir = default_config['image-dir']
        self.filename = os.path.join(self.image_dir, self.name + ".raw")

    def get_edges(self):
        """Because this is created without base, there are no edges."""
        return ([], [])

    def create(self):
        logger.debug("[%s] Creating loop on [%s] with size [%d]",
                     self.name, self.filename, self.size)

        self.add_rollback(image_delete, self.filename)
        image_create(self.filename, self.size)

        block_device = loopdev_attach(self.filename)
        self.add_rollback(loopdev_detach, block_device)

        if 'blockdev' not in self.state:
            self.state['blockdev'] = {}

        self.state['blockdev'][self.name] = {"device": block_device,
                                             "image": self.filename}
        logger.debug("Created loop  name [%s] device [%s] image [%s]",
                     self.name, block_device, self.filename)
        return

    def umount(self):
        loopdev_detach(self.state['blockdev'][self.name]['device'])

    def delete(self):
        image_delete(self.state['blockdev'][self.name]['image'])


class LocalLoop(PluginBase):

    def __init__(self, config, defaults, state):
        super(LocalLoop, self).__init__()
        self.node = LocalLoopNode(config, defaults, state)

    def get_nodes(self):
        return [self.node]



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import codecs
import fixtures
import json
import logging
import os

from stevedore import extension
from testtools.matchers import FileExists

import diskimage_builder.block_device.blockdevice as bd
import diskimage_builder.block_device.tests.test_base as tb

from diskimage_builder.block_device.exception import \
    BlockDeviceSetupException

logger = logging.getLogger(__name__)


class TestStateBase(tb.TestBase):

    def setUp(self):
        super(TestStateBase, self).setUp()

        # override the extensions to the test extensions
        test_extensions = extension.ExtensionManager(
            namespace='diskimage_builder.block_device.plugin_test',
            invoke_on_load=False)
        extensions_fixture = fixtures.MonkeyPatch(
            'diskimage_builder.block_device.config._extensions',
            test_extensions)
        self.useFixture(extensions_fixture)

        # status and other bits saved here
        self.build_dir = fixtures.TempDir()
        self.useFixture(self.build_dir)


class TestState(TestStateBase):

    # The the state generation & saving methods
    def test_state_create(self):
        params = {
            'build-dir': self.build_dir.path,
            'config': self.get_config_file('cmd_create.yaml')
        }

        bd_obj = bd.BlockDevice(params)

        bd_obj.cmd_init()
        bd_obj.cmd_create()

        # cmd_create should have persisted this to disk
        state_file = bd_obj.state_json_file_name
        self.assertThat(state_file, FileExists())

        # ensure we see the values put in by the test extensions
        # persisted
        with codecs.open(state_file, encoding='utf-8', mode='r') as fd:
            state = json.load(fd)
        self.assertDictEqual(state,
                             {'test_a': {'value': 'foo',
                                         'value2': 'bar'},
                              'test_b': {'value': 'baz'},
                              'test_init_state': 'here'})

        pickle_file = bd_obj.node_pickle_file_name
        self.assertThat(pickle_file, FileExists())

        # run umount, which should load the picked nodes and run in
        # reverse.  This will create some state in "test_b" that it
        # added to by "test_a" ... ensuring it was run backwards.  It
        # also checks the state was persisted through the pickling
        # process.
        bd_obj.cmd_umount()

    # Test state going missing between phases
    def test_missing_state(self):
        params = {
            'build-dir': self.build_dir.path,
            'config': self.get_config_file('cmd_create.yaml')
        }

        bd_obj = bd.BlockDevice(params)
        bd_obj.cmd_init()
        bd_obj.cmd_create()

        # cmd_create should have persisted this to disk
        state_file = bd_obj.state_json_file_name
        self.assertThat(state_file, FileExists())
        pickle_file = bd_obj.node_pickle_file_name
        self.assertThat(pickle_file, FileExists())

        # simulate the state somehow going missing, and ensure that
        # later calls notice
        os.unlink(state_file)
        os.unlink(pickle_file)
        # This reads from the state dump json file
        self.assertRaisesRegex(BlockDeviceSetupException,
                               "State dump not found",
                               bd_obj.cmd_getval, 'image-path')
        self.assertRaisesRegex(BlockDeviceSetupException,
                               "State dump not found",
                               bd_obj.cmd_writefstab)

        # this uses the pickled nodes
        self.assertRaisesRegex(BlockDeviceSetupException,
                               "Pickle file not found",
                               bd_obj.cmd_delete)
        self.assertRaisesRegex(BlockDeviceSetupException,
                               "Pickle file not found",
                               bd_obj.cmd_cleanup)

        # XXX: figure out unit test for umount

    # Test ordering of rollback calls if create() fails
    def test_rollback(self):
        params = {
            'build-dir': self.build_dir.path,
            'config': self.get_config_file('rollback.yaml'),
            'test_rollback': True
        }

        bd_obj = bd.BlockDevice(params)
        bd_obj.cmd_init()

        # The config file has flags in that tell the last node to
        # fail, which will trigger the rollback.
        self.assertRaises(RuntimeError, bd_obj.cmd_create)

        # cmd_create should have persisted this to disk even after the
        # failure
        state_file = bd_obj.state_json_file_name
        self.assertThat(state_file, FileExists())
        with codecs.open(state_file, encoding='utf-8', mode='r') as fd:
            state = json.load(fd)

        # ensure the rollback was called in order
        self.assertListEqual(state['rollback_test'],
                             ['never', 'gonna', 'give', 'you', 'up',
                              'never', 'gonna', 'let', 'you', 'down'])



#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import abc
import logging

#
# Plugins convert configuration entries into graph nodes ready for
# processing.  This defines the abstract classes for both.
#

logger = logging.getLogger(__name__)


class NodeBase(object, metaclass=abc.ABCMeta):
    """A configuration node entry

    This is the main driver class for dib-block-device operation.

    The final operations graph is composed of instantiations of this
    class.  The graph undergoes a topological sort (i.e. is linearised
    in dependency order) and each node has :func:`create` called in
    order to perform its operations.

    Every node has a unique string ``name``.  This is its key in the
    graph and used for edge relationships.  Implementations must
    ensure they initalize it; e.g.

    .. code-block:: python

       class FooNode(NodeBase):
           def __init__(name, arg1, ...):
               super(FooNode, self).__init__(name)

    """
    def __init__(self, name, state):
        self.name = name
        self.state = state
        self.rollbacks = []

    def get_name(self):
        return self.name

    def add_rollback(self, func, *args, **kwargs):
        """Add a call for rollback

        Functions registered with this method will be called in
        reverse-order in the case of failures during
        :func:`Nodebase.create`.

        :param func: function to call
        :param args: arguments
        :param kwargs: keyword arguments
        :return: None
        """
        self.rollbacks.append((func, args, kwargs))

    def rollback(self):
        """Initiate rollback

        Call registered rollback in reverse order.  This method is
        called by the driver in the case of failures during
        :func:`Nodebase.create`.

        :return None:
        """
        # XXX: maybe ignore SystemExit so we always continue?
        logger.debug("Calling rollback for %s", self.name)
        for func, args, kwargs in reversed(self.rollbacks):
            func(*args, **kwargs)

    @abc.abstractmethod
    def get_edges(self):

        """Return the dependencies/edges for this node

        This function will be called after all nodes are created (this
        is because some plugins need to know the global state of all
        nodes to decide their dependencies).

        This function returns a tuple with two lists

         * ``edges_from`` : a list of node names that point to us
         * ``edges_to`` : a list of node names we point to

        In most cases, node creation will have saved a single parent
        that was given in the ``base`` parameter of the configuration.
        A usual return might look like:

        .. code-block:: python

           def get_edges(self):
               return ( [self.base], [] )

        Some nodes (``level0``) don't have a base, however
        """
        return

    @abc.abstractmethod
    def create(self):
        """Main creation driver

        This is the main driver function.  After the graph is
        linearised, each node has it's :func:`create` function called.

        :raises Exception: A failure should raise an exception.  This
          will initiate a rollback.  See :func:`Nodebase.add_rollback`.
        :return: None
        """
        return

    def umount(self):
        """Umount actions

        Actions to taken when ``dib-block-device umount`` is called.
        The nodes are called in the reverse order to :func:`create`

        :return: None

        """
        return

    def cleanup(self):
        """Cleanup actions

        Actions to taken when ``dib-block-device cleanup`` is called.
        This is the cleanup path in the *success* case.  The nodes are
        called in the reverse order to :func:`create`

        :return: None
        """
        return

    def delete(self):
        """Cleanup actions

        Actions to taken when ``dib-block-device delete`` is called.
        This is the cleanup path in case of a reported external
        *failure*.  The nodes are called in the reverse order to
        :func:`create`

        :return: None
        """
        return


class PluginBase(object, metaclass=abc.ABCMeta):
    """The base plugin object

    This is the base plugin object.  Plugins are an instantiation of
    this class.  There should be an entry-point (see setup.cfg)
    defined under ``diskimage_builder.block_device.plugin`` for each
    plugin, e.g.

      foo = diskimage_builder.block_device.levelX.foo:Foo

    A configuration entry in the graph config that matches this entry
    point will create an instance of this class, e.g.

    .. code-block:: yaml

       foo:
         name: foo_node
         base: parent_node
         argument_a: bar
         argument_b: baz

    The ``__init__`` function will be passed three arguments:

    ``config``
       The full configuration dictionary for the entry.
       A unique ``name`` entry can be assumed.  In most cases
       a ``base`` entry will be present giving the parent node
       (see :func:`NodeBase.get_edges`).
    ``state``
       A reference to the gobal state dictionary.  This should be
       passed to :func:`NodeBase.__init__` on node creation
    ``defaults``
       The global defaults dictionary (see ``--params``)

    ``get_nodes()`` should return the node object(s) created by the
    config for insertion into the final configuration graph.  In the
    simplest case, this is probably a single node created during
    instantiation.  e.g.

    .. code-block:: python

       class Foo(PluginBase):

         def __init__(self, config, defaults, state):
             super(Foo, self).__init__()
             self.node = FooNode(config.name, state, ...)

         def get_nodes(self):
             return [self.node]


    Some plugins require more, however.
    """
    def __init__(self):
        pass

    @abc.abstractmethod
    def get_nodes(self):
        """Return nodes created by the plugin

        :returns: a list of :class:`.NodeBase` objects for insertion
          into the graph
        """
        return
